/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow;

import com.google.inject.Injector;
import org.zdevra.sparrow.actor.ActorExceptionHandler;
import org.zdevra.sparrow.actor.ActorHandler;
import org.zdevra.sparrow.actor.PublisherSubsriberActor;
import org.zdevra.sparrow.actor.StandaloneActor;
import org.zdevra.sparrow.channel.MessageChannel;
import org.zdevra.sparrow.channel.SubscribableChannel;
import org.zdevra.sparrow.mailbox.Mailbox;

import javax.inject.Inject;

/**
 * Extended {@link org.zdevra.sparrow.actor.PublisherSubsriberActor actor} where during initialization phase are objects like input/output channel etc injected
 * from guice.
 *
 * @author Zdenko Vrabel (zdenko.vrabel@celum.com)
 */
class GuiceActor extends StandaloneActor
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** class of actor handler. Google guice creates instance for this class */
	private final Class<? extends ActorHandler> actorHandlerClass;

	/** guice injector */
	@Inject private Injector injector;

	//-------------------------------------------------------------------------------------------------------------------
	// constructors
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Constructor
	 *
	 * @param handlerClass
	 * @param mailbox
	 * @param exceptionHandler
	 * @param inputName
	 * @param outputName
	 */
	GuiceActor(Class<? extends ActorHandler>  handlerClass, Mailbox mailbox, ActorExceptionHandler exceptionHandler, String inputName, String outputName )
	{
		super(null, mailbox, exceptionHandler, inputName, outputName);
		this.actorHandlerClass = handlerClass;
	}


	/**
	 * Constructor
	 *
	 * @param handlerClass
	 * @param mailbox
	 * @param exceptionHandler
	 * @param input
	 * @param output
	 */
	GuiceActor(Class<? extends ActorHandler>  handlerClass, Mailbox mailbox, ActorExceptionHandler exceptionHandler, SubscribableChannel input, MessageChannel output)
	{
		super(null, mailbox, exceptionHandler, input, output);
		this.actorHandlerClass = handlerClass;
	}


	/**
	 * Constructor
	 *
	 * @param handler
	 * @param mailbox
	 * @param exceptionHandler
	 * @param inputName
	 * @param outputName
	 */
	GuiceActor(ActorHandler<?> handler, Mailbox mailbox, ActorExceptionHandler exceptionHandler, String inputName, String outputName )
	{
		super(handler, mailbox, exceptionHandler, inputName, outputName);
		this.actorHandlerClass = null;
	}


	/**
	 * Constructor
	 *
	 * @param handler
	 * @param mailbox
	 * @param exceptionHandler
	 * @param input
	 * @param output
	 */
	GuiceActor(ActorHandler<?> handler, Mailbox mailbox, ActorExceptionHandler exceptionHandler, SubscribableChannel input, MessageChannel output)
	{
		super(handler, mailbox, exceptionHandler, input, output);
		this.actorHandlerClass = null;
	}


	//-------------------------------------------------------------------------------------------------------------------
	// impl. methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Implementation does injection of actor handler from class.
	 *
	 * {@inheritDoc}
	 */
	@Override
	public void onInit(SparrowEngine engine)
	{
		if (actorHandler == null) {
			actorHandler = injector.getInstance(actorHandlerClass);
		} else {
			injector.injectMembers(actorHandler);
		}

		super.onInit(engine);
	}

}
