/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow;

import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.Key;
import org.zdevra.sparrow.channel.ChannelNames;
import org.zdevra.sparrow.channel.ChannelReference;
import org.zdevra.sparrow.channel.MessageChannel;

import java.util.Set;

/**
 * Engine class start and shutdown all entities of sparrow like actors etc. Class is used by {@link SparrowModule}
 * because members of Guice Modules cannot be injected.
 *
 * @author Zdenko Vrabel (vrabel.zdenko@gmail.com)
 */
class GuiceSparrowEngine implements SparrowEngine
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** all units implementing {@link SparrowLifecycle} injected from guice via multibinding */
	@Inject private volatile Set<SparrowLifecycle> units;

	/** guice injector */
	@Inject private Injector injector;



	//-------------------------------------------------------------------------------------------------------------------
	// methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void start()
	{
		//initialize all units
		for (SparrowLifecycle unit: units)
		{
			unit.onInit(this);
		}

		//start all units
		for (SparrowLifecycle unit: units)
		{
			unit.onStart();
		}
	}


	/**
	 * Shutdown the sparrow engine and all his units
	 */
	@Override
	public void shutdown() throws InterruptedException
	{
		for (SparrowLifecycle unit: units)
		{
			unit.onStop();
		}
	}


	/**
	 * Method returns the {@link MessageChannel channel} instance by given name
	 *
	 * @param channelRef
	 *          reference fo channel (see {@link org.zdevra.sparrow.channel.NamedChannelReference})
	 *
	 * @return instance of channel. If there is no channel for given name, then method returns null.
	 */
	@Override
	public MessageChannel getMessageChannel(ChannelReference channelRef)
	{
		if (channelRef != null && channelRef.name() != null) {

			return injector.getInstance(Key.get(MessageChannel.class, ChannelNames.named(channelRef.name())));
		}
		return null;
	}

}
