/*****************************************************************************
 * Copyright 2012 Zdenko Vrabel
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package org.zdevra.sparrow;

import com.google.inject.PrivateModule;
import com.google.inject.multibindings.Multibinder;
import org.zdevra.sparrow.actor.*;
import org.zdevra.sparrow.adapter.AbstractAdapterBuilder;
import org.zdevra.sparrow.adapter.AdapterBuilder;
import org.zdevra.sparrow.adapter.InboundAdapter;
import org.zdevra.sparrow.aggregator.AggregatorBuilder;
import org.zdevra.sparrow.aggregator.AggregatorBuilderImpl;
import org.zdevra.sparrow.chain.ChainBuilder;
import org.zdevra.sparrow.channel.*;
import org.zdevra.sparrow.dsl.FromDsl;
import org.zdevra.sparrow.dsl.SparrowDsl;
import org.zdevra.sparrow.filter.FilterBuilder;
import org.zdevra.sparrow.filter.HeaderFilterRuleBuilder;
import org.zdevra.sparrow.filter.PayloadFilterRuleBuilder;
import org.zdevra.sparrow.smartproxy.BaseSmartProxyBuilder;
import org.zdevra.sparrow.smartproxy.SmartProxyBuilder;
import org.zdevra.sparrow.splitter.SplitterBuilder;
import org.zdevra.sparrow.splitter.SplitterBuilderImpl;

import java.util.LinkedList;
import java.util.List;

/**
 * A module who initialize and construct message system. You will implement 'configureSparrow' for determining your
 * message system.
 *
 * @author Zdenko Vrabel (vrabel.zdenko@gmail.com)
 */
public abstract class SparrowModule extends PrivateModule implements SparrowEngine, SparrowDsl
{
	//-------------------------------------------------------------------------------------------------------------------
	// members
	//-------------------------------------------------------------------------------------------------------------------

	/** engine starts/shutdown the sparrow units */
	private SparrowEngine engine;

	/** multibinder for all sparrow units */
	private Multibinder<SparrowLifecycle> sparrowMultibinder;

	/** list of all pre builders. It's temporary and exists only during 'configureSparrow' call. */
	private List<SparrowBuilder> preBuilders;

	/** list of all channel builders. It's temporary and exists only during 'configureSparrow' call. */
	private List<ChannelBuilder> channelBuilders;

	/** list of all actor builders. It's temporary and exists only during 'configureSparrow' call. */
	private List<ActorBuilder> actorBuilders;


	//-------------------------------------------------------------------------------------------------------------------
	// impl. & abstract methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Here you will implement code which determines your message system.
	 */
	protected abstract void configureSparrow();


	@Override
	protected final void configure()
	{
		try {
			sparrowMultibinder = Multibinder.newSetBinder(binder(), SparrowLifecycle.class);
			preBuilders = new LinkedList<SparrowBuilder>();
			channelBuilders = new LinkedList<ChannelBuilder>();
			actorBuilders = new LinkedList<ActorBuilder>();

			configureSparrow();

			for (SparrowBuilder preBuilder : preBuilders) {
				preBuilder.build();
			}

			for (ChannelBuilder channelBuilder : channelBuilders) {
				channelBuilder.build();
			}

			for (ActorBuilder actorBuilder : actorBuilders) {
				actorBuilder.build();
			}

			this.engine = new GuiceSparrowEngine();
			bind(SparrowEngine.class).toInstance(engine);

		} finally {
			sparrowMultibinder = null;
			channelBuilders = null;
			actorBuilders = null;
			preBuilders = null;
		}
	}


	//-------------------------------------------------------------------------------------------------------------------
	// builders impl.
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Guice impl. of {@link org.zdevra.sparrow.adapter.AdapterBuilder}
	 */
	private class GuiceAdapterBuilder extends AbstractAdapterBuilder
	{
		/**
		 * {@inheritDoc}
		 */
		@Override
		public void asInstance(InboundAdapter adapter)
		{
			adapter.setOutputName(this.outputName);
			sparrowMultibinder.addBinding().toInstance(adapter);
		}


		/**
		 * {@inheritDoc}
		 */
		@Override
		public void asClass(Class<? extends InboundAdapter> adapterClass)
		{
			sparrowMultibinder.addBinding().to(adapterClass).asEagerSingleton();
		}
	}



	/**
	 * Guice impl. of {@link ChannelBuilder}
	 */
	private class GuiceChannelBuilder extends AbstractChannelBuilder
	{
		/** flag determine whether channel is exposed from private module (Guice) */
		private boolean exposed = false;

		private Class<? extends MessageChannel> channelClass;

		/**
		 * {@inheritDoc}
		 */
		@Override
		public final ChannelBuilder exposed()
		{
			this.exposed = true;
			return this;
		}


		@Override
		public ChannelBuilder asClass(Class<? extends MessageChannel> channelClass)
		{
			this.channelClass = channelClass;
			return this;
		}


		@Override
		public void build()
		{
			if (channelClass != null) {
				bind(MessageChannel.class)
						  .annotatedWith(ChannelNames.named(this.channelName))
						  .toProvider(new ChannelProvider(channelClass, channelName))
						  .asEagerSingleton();
			}
			else if (channelInstance != null) {
				bind(MessageChannel.class)
						  .annotatedWith(ChannelNames.named(this.channelName))
						  .toInstance(channelInstance);

			} else {
				//default channel
				bind(MessageChannel.class)
						  .annotatedWith(ChannelNames.named(this.channelName))
						  .toProvider(new ChannelProvider(DirectChannel.class, channelName))
						  .asEagerSingleton();
			}

			if (this.exposed) {
				SparrowModule.this.expose(MessageChannel.class).annotatedWith(ChannelNames.named(channelName));
			}
		}
	}


	/**
	 * Guice builder impl. for {@link org.zdevra.sparrow.dsl.FromDsl}
	 */
	private class GuiceFromBuilderImpl extends FromDsl
	{
		@Override
		public ChainBuilder chain(String name)
		{
			ChainBuilder chainBuilder = new ChainBuilder(this.inputChannelRef, name, SparrowModule.this);
			preBuilders.add(chainBuilder);
			return chainBuilder;
		}


		@Override
		public ActorBuilder actor(ActorHandler<?> actor)
		{
			return actor(new InstanceActorHandlerBuilder(actor));
		}


		@Override
		public ActorBuilder actor(ActorHandlerBuilder actor)
		{
			GuiceActorBuilder b = new GuiceActorBuilder(actor);
			b.from(this.inputChannelRef);
			return b;
		}


		@Override
		public ActorBuilder actor(Class<? extends ActorHandler> actor)
		{
			GuiceActorBuilder b = new GuiceActorBuilder(actor);
			b.from(this.inputChannelRef);
			return b;
		}
	}


	/**
	 * Actor {@link ActorBuilder builder} impl. for Guice
	 */
	private class GuiceActorBuilder extends AbstractActorBuilder
	{
		private Class<? extends ActorHandler> handlerClass;


		private GuiceActorBuilder(ActorHandlerBuilder handlerBuilder)
		{
			super(handlerBuilder);
			actorBuilders.add(this);
		}


		private GuiceActorBuilder(Class<? extends ActorHandler> handlerClass)
		{
			super(null);
			actorBuilders.add(this);
			this.handlerClass = handlerClass;
		}


		@Override
		public void build()
		{
			AbstractActor actorInstance = null;

			if (handlerClass != null) {
				actorInstance =
					new GuiceActor(handlerClass,
							  this.mailbox,
							  this.exceptionHandler,
							  this.inputChannel,
							  this.outputChannel);

				SparrowModule.this.bind(handlerClass);

				if (SparrowLifecycle.class.isAssignableFrom(handlerClass)) {
					sparrowMultibinder.addBinding().to((Class<? extends SparrowLifecycle>) handlerClass);
				}

			} else {
				ActorHandler<?> handler = this.handlerBuilder.build();
				actorInstance =
					  new GuiceActor(this.handlerBuilder.build(),
								 this.mailbox,
								 this.exceptionHandler,
								 this.inputChannel,
								 this.outputChannel);

				if (handler instanceof SparrowLifecycle) {
					sparrowMultibinder.addBinding().toInstance((SparrowLifecycle) handler);
				}
			}

			actorInstance.setName(this.actorName);
			SparrowModule.this.bindActor(actorInstance, this.actorName);
		}
	}


	/**
	 * does binding of actor instance into Guice
	 *
	 * @param actorInstance
	 * @param toName
	 */
	private void bindActor(AbstractActor actorInstance, String toName)
	{
		bind(MessageChannel.class)
				.annotatedWith(ChannelNames.named(toName))
				.toInstance(actorInstance);

		sparrowMultibinder.addBinding().toInstance(actorInstance);
	}


	//-------------------------------------------------------------------------------------------------------------------
	// public methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void start()
	{
		engine.start();
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public void shutdown() throws InterruptedException
	{
		engine.shutdown();
	}


	/**
	 * {@inheritDoc}
	 */
	@Override
	public MessageChannel getMessageChannel(ChannelReference channelName)
	{
		return engine.getMessageChannel(channelName);
	}



	//-------------------------------------------------------------------------------------------------------------------
	// DSL methods
	//-------------------------------------------------------------------------------------------------------------------

	/**
	 * Method returns reference to channel following entered named
	 */
	@Override
	public ChannelReference channelRef(String name)
	{
		return new NamedChannelReference(name);
	}


	/**
	 * Method create new channel and naming the channel with some name. You could use named channel
	 * with {@link org.zdevra.sparrow.channel.ChannelNamed} annotation which ensure injecting the channel.
	 */
	@Override
	public ChannelBuilder channel(String channelName)
	{
		GuiceChannelBuilder b = new GuiceChannelBuilder();
		b.namedAs(channelName);
		channelBuilders.add(b);
		return b;
	}


	/**
	 * Method build a new actor. InputOutputActor is running in one thread and process messages in
	 * {@link org.zdevra.sparrow.actor.ActorHandler} implementation.
	 *
	 * @return
	 */
	@Override
	public ActorBuilder actor(ActorHandler<?> actor)
	{
		ActorBuilder builder = new GuiceActorBuilder(new InstanceActorHandlerBuilder(actor));
		return builder;
	}


	/**
	 * Method build a new actor. InputOutputActor is running in one thread and process messages in
	 * {@link org.zdevra.sparrow.actor.ActorHandler} implementation.
	 *
	 * @return
	 */
	@Override
	public ActorBuilder actor(ActorHandlerBuilder actor)
	{
		ActorBuilder builder = new GuiceActorBuilder(actor);
		return builder;
	}


	/**
	 * Method build a new actor. InputOutputActor is running in one thread and process messages in
	 * {@link org.zdevra.sparrow.actor.ActorHandler} implementation.
	 *
	 * @return
	 */
	@Override
	public ActorBuilder actor(Class<? extends ActorHandler> actor)
	{
		ActorBuilder builder = new GuiceActorBuilder(actor);
		return builder;
	}


	/**
	 * build and register inbound {@link InboundAdapter adapter}.
	 *
	 */
	@Override
	public AdapterBuilder inboundAdapter()
	{
		AdapterBuilder builder = new GuiceAdapterBuilder();
		return builder;
	}


	/**
	 * Create/register smart {@link org.zdevra.sparrow.smartproxy.SmartProxy proxy} which is more complex entity.
	 *
	 * @param name
	 *          name for smart proxy
	 */
	@Override
	public SmartProxyBuilder smartProxy(String name)
	{
		return new BaseSmartProxyBuilder(name, this);
	}


	@Override
	public AggregatorBuilder aggregator()
	{
		return new AggregatorBuilderImpl();
	}


	@Override
	public SplitterBuilder splitter()
	{
		return new SplitterBuilderImpl();
	}


	@Override
	public FilterBuilder filter()
	{
		return new FilterBuilder();
	}


	@Override
	public HeaderFilterRuleBuilder header(String header)
	{
		return new HeaderFilterRuleBuilder(header);
	}


	@Override
	public PayloadFilterRuleBuilder payload()
	{
		return new PayloadFilterRuleBuilder();
	}


	@Override
	public FromDsl from(ChannelReference channel)
	{
		FromDsl fromBuilder = new GuiceFromBuilderImpl();
		fromBuilder.setInputChannel(channel);
		return fromBuilder;
	}
}
